import 'package:bird/fp.dart';

extension KindEitherFix<L, R> on Kind<Kind<ForEither, L>, R> {
  Either<L, R> get D => this as Either<L, R>;
}

class ForEither {
  const ForEither._();
}

class PartEither<A, B> implements Kind<ForEither, A>, Monad<PartEither, B> {
  const PartEither._();

  Either<A, B> get self => this as Either<A, B>;

  @override
  Kind<PartEither, C> bind<C>(Kind<PartEither, C> Function(B) f) =>
      self.bind((a) => f(a) as Either<A, C>);

  @override
  Kind<PartEither, C> ap<C>(Kind<PartEither, C Function(B)> f) =>
      self.ap(f as Either<A, C Function(B)>);

  @override
  Kind<PartEither, B> pure(B a) => self.pure(a);

  @override
  Kind<PartEither, C> map<C>(C Function(B) f) => self.map(f);
}

abstract class Either<A, B> implements Kind2<PartEither<A, B>, B, A>, Monad<Kind<ForEither, A>, B> {
  const Either._();

  T fold<T>(T Function(A l) ifLeft, T Function(B r) ifRight);

  void forEach(void Function(B) f) => fold((_) {}, f);

  @override
  String toString() => fold((a) => 'Left($a)', (a) => 'Right($a)');

  @override
  Either<A, C> map<C>(C Function(B) f) => bind((B a) => right(f(a)));

  @override
  Either<A, B> pure(B a) => right(a);

  @override
  Either<A, C> ap<C>(Kind<Kind<ForEither, A>, C Function(B)> f) =>
      bind((a) => KindEitherFix(f).D.map((b) => b(a)));

  @override
  Either<A, C> bind<C>(Kind<Kind<ForEither, A>, C> Function(B f) f) =>
      fold(left, (a) => KindEitherFix(f(a)).D);
}

class Left<L, R> extends Either<L, R> {
  final L _l;

  const Left(this._l) : super._();

  L get value => _l;

  @override
  B fold<B>(B Function(L l) ifLeft, B Function(R r) ifRight) => ifLeft(_l);

  @override
  bool operator ==(Object other) => other is Left && other._l == _l;

  @override
  int get hashCode => _l.hashCode;
}

class Right<L, R> extends Either<L, R> {
  final R _r;

  const Right(this._r) : super._();

  R get value => _r;

  @override
  B fold<B>(B Function(L l) ifLeft, B Function(R r) ifRight) => ifRight(_r);

  @override
  bool operator ==(Object other) => other is Right && other._r == _r;

  @override
  int get hashCode => _r.hashCode;
}
