import 'package:bird/fp.dart';

extension KindOptionFix<A> on Kind<ForOption, A> {
  Option<A> get D => this as Option<A>;
}

class ForOption<B> implements Kind<ForOption, B>, Monad<ForOption, B> {
  const ForOption._();

  @override
  Kind<ForOption, C> bind<C>(Kind<ForOption, C> Function(B) f) =>
      KindOptionFix(this).D.bind((a) => KindOptionFix(f(a)).D);

  @override
  Kind<ForOption, C> ap<C>(Kind<ForOption, C Function(B)> f) =>
      KindOptionFix(this).D.ap(KindOptionFix(f).D);

  @override
  Kind<ForOption, B> pure(B a) => KindOptionFix(this).D.pure(a);

  @override
  Kind<ForOption, C> map<C>(C Function(B) f) => KindOptionFix(this).D.map(f);
}

abstract class Option<A> implements Kind<ForOption, A>, Monad<ForOption, A> {
  const Option._();

  B fold<B>(B Function() none, B Function(A) some);

  void forEach(void Function(A) f) => fold(() {}, f);

  @override
  String toString() => fold(() => 'None', (a) => 'Some($a)');

  @override
  Option<B> map<B>(B Function(A) f) => bind((a) => some(f(a)));

  @override
  Option<A> pure(A a) => some<A>(a);

  @override
  Option<B> ap<B>(Kind<ForOption, B Function(A)> f) =>
      bind((a) => KindOptionFix(f).D.map((b) => b(a)));

  @override
  Option<B> bind<B>(Kind<ForOption, B> Function(A) f) =>
      fold(none, (a) => KindOptionFix(f(a)).D);
}

class Some<A> extends Option<A> {
  final A _a;

  const Some(this._a) : super._();

  @override
  B fold<B>(B Function() none, B Function(A) some) => some(_a);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is Some &&
              _a == other._a;

  @override
  int get hashCode => _a.hashCode;
}

class None<A> extends Option<A> {
  const None() : super._();

  @override
  B fold<B>(B Function() none, B Function(A) some) => none();

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is None;

  @override
  int get hashCode => 0;
}
