import 'package:bird/fp.dart';

abstract class Glass<GIn, GOut, C, SIn, SOut> {
  GOut get(GIn gin);

  SOut set(C c, SIn sin);
}

/// A more generalised polymorphic Lens with a custom setter companion.
class ConcreteGlass<GIn, GOut, C, SIn, SOut>
    implements Glass<GIn, GOut, C, SIn, SOut> {
  final GOut Function(GIn) _get;

  final SOut Function(C, SIn) _set;

  const ConcreteGlass(this._get, this._set);

  @override
  GOut get(GIn gin) => _get(gin);

  @override
  SOut set(C c, SIn sin) => _set(c, sin);
}

/// Use this to create new lenses.
GlassBlower lens = const GlassBlower._();

class GlassBlower {
  const GlassBlower._();

  /// A (monomorphic) lens.
  Glass<S, T, S, T, S> s<S, T>(T Function(S) get, S Function(S, T) set) {
    return ConcreteGlass(get, set);
  }

  Glass<Kind<ForId, S>,
      Kind<ForId, T>,
      Kind<ForId, S>,
      Kind<ForId, T>,
      Kind<ForId, S>> id<S, T>(T Function(S) get,
      S Function(S, T) set) => ConcreteGlass(get, set).toId;

  Glass<S, Kind<F, T>, S, T, Kind<F, S>> m<S, T, F>(Kind<F, T> Function(S) get,
      Kind<F, S> Function(S, T) set) {
    return ConcreteGlass(get, set);
  }

  Glass<S, Kind<F, T>, S, T, S> lm<S, T, F>(Kind<F, T> Function(S) get,
      S Function(S, T) set) {
    return ConcreteGlass(get, set);
  }

  Glass<S, T, S, T, Kind<F, S>> rm<S, T, F>(T Function(S) get,
      Kind<F, S> Function(S, T) set) {
    return ConcreteGlass(get, set);
  }

  Glass<void, Kind<ForId, GIn>, void, Kind<ForId, GIn>, void> prop<GIn>(
      GIn Function() get,
      void Function(GIn) set) {
    return ConcreteGlass<void, Kind<ForId, GIn>, void, Kind<ForId, GIn>, void>(
          (_) => Id(get()),
          (_, a) => a.D.map<void>(set),
    );
  }
}
