import 'package:bird/fp.dart';

Either<L, R> left<L, R>(L v) => Left(v);

Either<L, R> right<L, R>(R v) => Right(v);

Either<MapEntry<dynamic, StackTrace>, A> eitherTry<A>(A Function() thunk) {
  try {
    return right(thunk());
  } catch (e, s) {
    return left(MapEntry<dynamic, StackTrace>(e, s));
  }
}

extension IEitherExtension<L, R> on Either<L, R> {
  Either<L, R> orElse(Either<L, R> Function() other) {
    return fold((_) {
      return other();
    }, (_) {
      return this;
    });
  }

  R getOrElse(R Function() dflt) {
    return fold((_) {
      return dflt();
    }, (a) {
      return a;
    });
  }

  R operator |(R dflt) {
    return getOrElse(() {
      return dflt;
    });
  }

  Either<L2, R> leftMap<L2>(L2 Function(L l) f) {
    return fold((L l) {
      return left(f(l));
    }, right);
  }

  Option<R> toOption() {
    return fold(none, some);
  }

  Option<L> toLeftOption() {
    return fold(some, (_) => none());
  }

  bool isLeft() {
    return fold((_) => true, (_) => false);
  }

  bool isRight() {
    return !isLeft();
  }

  Either<R, L> swap() {
    return fold(right, left);
  }
}
