import 'package:bird/fp.dart';
import 'package:dartz/dartz.dart' show IList, nil, cons;

extension ListOption<T> on Iterable<T> {
  Option<T> get firstOrNone {
    return this.isEmpty ? none() : some(this.first);
  }

  Option<T> firstWhereOrNone(bool Function(T) predicate) {
    return optionOf(this.firstWhere(predicate, orElse: () => null));
  }

  Option<T> get lastOrNone {
    return this.isEmpty ? none() : some(this.last);
  }

  Option<T> lastWhereOrNone(bool Function(T) predicate) {
    return optionOf(this.lastWhere(predicate, orElse: () => null));
  }

  Option<T> get singleOrNone {
    return this.length == 1 ? some(this.first) : some(this.last);
  }

  Option<T> singleWereOrNone(bool Function(T) predicate) {
    return ListOption(where(predicate)).singleOrNone;
  }
}

extension IListTraverseEither2<A> on IList<A>{
  Either<L, IList<B>> traverseEither2<B, L>(Either<L, B> Function(A a) f) {
    return foldLeft<Either<L, IList<B>>>(right(nil()),
          (a, gb) => a.bind((a) => f(gb).map((h) => cons(h, a))),
    ).map((l) => l.reverse());
  }
}

