import 'package:bird/fp.dart';

Option<A> none<A>([dynamic _]) => None<A>();

Option<A> some<A>(A a) => Some(a);

Option<A> optionOf<A>(A a) => a != null ? Some<A>(a) : None<A>();

Option<List<A>> optionOfList<A>(List<A> i) {
  return (i.isEmpty) ? none() : some(i);
}

Option<Iterable<A>> optionOfIterable<A>(Iterable<A> i) {
  return (i.isEmpty) ? none() : some(i);
}

extension OptionExtensions<A> on Option<A> {
  Option<A> orElse(Option<A> Function() defaultValue) {
    return fold(defaultValue, some);
  }

  List<A> toList() {
    return fold(() => [], (a) => [a]);
  }

  A getOrElse(A Function() dflt) {
    return fold(dflt, (a) {
      return a;
    });
  }

  A operator |(A dflt) {
    return getOrElse(() => dflt);
  }

  bool isSome() {
    return fold(() => false, (_) => true);
  }

  bool isNone() {
    return !isSome();
  }

  Tuple2<Option<A>, Option<B>> and<B>(Option<B> b) {
    return Tuple2(this, b);
  }
}
