import 'package:bird/fp.dart';

extension TwoOption<A, B> on Tuple2<Option<A>, Option<B>> {
  Option<Z> map<Z>(Z Function(A, B) map) {
    return value1.bind((a) {
      return value2.bind((b) {
        return some(map(a, b));
      });
    });
  }

  Tuple3<Option<A>, Option<B>, Option<Z>> and<Z>(Option<Z> other) =>
      Tuple3(value1, value2, other);
}

extension ThreeOption<A, B, C> on Tuple3<Option<A>, Option<B>, Option<C>> {
  Option<Z> map<Z>(Z Function(A, B, C) map) {
    return a.bind((a) {
      return b.bind((b) {
        return c.bind((c) {
          return some(map(a, b, c));
        });
      });
    });
  }

  Tuple4<Option<A>, Option<B>, Option<C>, Option<Z>> and<Z>(Option<Z> other) =>
      Tuple4(value1, value2, value3, other);
}

extension FourOption<A, B, C, D> on Tuple4<Option<A>,
    Option<B>,
    Option<C>,
    Option<D>>{

  Option<Z> map<Z>(Z Function(A, B, C, D) map) {
    return value1.bind((a) {
      return value2.bind((b) {
        return value3.bind((c) {
          return value4.bind((d) {
            return some(map(a, b, c, d));
          });
        });
      });
    });
  }

  Tuple5<Option<A>, Option<B>, Option<C>, Option<D>, Option<Z>> and<Z>(
      Option<Z> other) => Tuple5(value1, value2, value3, value4, other);
}

extension FiveOption<A, B, C, D, E> on Tuple5<Option<A>,
    Option<B>,
    Option<C>,
    Option<D>,
    Option<E>> {
  Option<Z> map<Z>(Z Function(A, B, C, D, E) map) {
    return a.bind((a) {
      return b.bind((b) {
        return c.bind((c) {
          return d.bind((d) {
            return e.bind((e) => some(map(a, b, c, d, e)));
          });
        });
      });
    });
  }

  Tuple6<Option<A>,
      Option<B>,
      Option<C>,
      Option<D>,
      Option<E>,
      Option<Z>> and<Z>(Option<Z> other) => Tuple6(a, b, c, d, e, other);
}

extension SixOption<A, B, C, D, E, F> on Tuple6<Option<A>,
    Option<B>,
    Option<C>,
    Option<D>,
    Option<E>,
    Option<F>> {

  Option<Z> map<Z>(Z Function(A, B, C, D, E, F) map) {
    return a.bind((a) {
      return b.bind((b) {
        return c.bind((c) {
          return d.bind((d) {
            return e.bind((e) {
              return f.bind((f) => some(map(a, b, c, d, e, f)));
            });
          });
        });
      });
    });
  }

  Tuple7<Option<A>,
      Option<B>,
      Option<C>,
      Option<D>,
      Option<E>,
      Option<F>,
      Option<Z>> and<Z>(Option<Z> other) =>
      Tuple7(
          a,
          b,
          c,
          d,
          e,
          f,
          other);
}
