import 'package:bird/fp.dart';
import 'package:test/test.dart';

void main() {
  group("Either", () {
    test("fold", () {
      final a = left<String, int>("x").fold((a) => a, (a) =>
      throw Exception(""));
      expect(a, "x");
      final b = right<String, int>(1).fold((a) => throw Exception(""), (
          a) => a);
      expect(b, 1);
    });
    test("forEach", () {
      int a;
      left<String, int>("x").forEach((f) => a = f);
      expect(a, null);
      int b;
      right<String, int>(1).forEach((f) => b = f);
      expect(b, 1);
    });
    test("toString", () {
      expect(left<String, int>("x").toString(), "Left(x)");
      expect(right<String, int>(1).toString(), "Right(1)");
    });
    group("monad", () {
      test("map", () {
        expect(left<String, int>("x").map(times(2)), left<String, int>("x"));
        expect(right<String, int>(1).map(times(2)), right<String, int>(2));
      });
      test("bind", () {
        expect(left<String, int>("x").map(times(2)), left<String, int>("x"));
        expect(right<String, int>(1).map(times(2)), right<String, int>(2));
      });
      test("ap", () {
        expect(left<String, int>("x").ap<int>(lTimes2), left<String, int>("x"));
        expect(right<String, int>(1).ap<int>(rTimes2), right<String, int>(2));
        expect(left<String, int>("x").ap<int>(rTimes2), left<String, int>("x"));
        expect(right<String, int>(1).ap<int>(lTimes2), left<String, int>("-"));
      });
      test("pure", () {
        expect(left<String, int>("x").pure(3), right<String, int>(3));
        expect(right<String, int>(1).pure(4), right<String, int>(4));
      });
      // laws
      // ...
    });
  });
}

int Function(int) times(int i) => (a) => a * i;

Either<String, int Function(int)> lTimes2 = left("-");

Either<String, int Function(int)> rTimes2 = right(times(2));
